import lab as B
import pytest

from stheno.input import Input, MultiInput
from stheno.kernel import EQ
from stheno.measure import Measure, GP
from stheno.mokernel import MultiOutputKernel
from .util import approx


@pytest.mark.parametrize("x1", [B.linspace(0, 1, 10), Input(B.linspace(0, 1, 10))])
@pytest.mark.parametrize("x2", [B.linspace(1, 2, 5), Input(B.linspace(0, 2, 5))])
@pytest.mark.parametrize("x3", [B.linspace(1, 2, 10), Input(B.linspace(1, 2, 10))])
def test_mokernel(x1, x2, x3):
    m = Measure()
    p1 = GP(1 * EQ(), measure=m)
    p2 = GP(2 * EQ().stretch(2), measure=m)

    k = MultiOutputKernel(m, p1, p2)
    ks = m.kernels

    # Check representation.
    assert str(k) == "MultiOutputKernel(EQ(), 2 * (EQ() > 2))"

    # Input versus input:
    approx(
        k(x1, x2),
        B.concat2d(
            [ks[p1, p1](x1, x2), ks[p1, p2](x1, x2)],
            [ks[p2, p1](x1, x2), ks[p2, p2](x1, x2)],
        ),
    )
    approx(
        k.elwise(x1, x3),
        B.concat(ks[p1, p1].elwise(x1, x3), ks[p2, p2].elwise(x1, x3), axis=0),
    )

    # Input versus `FDD`:
    approx(k(p1(x1), x2), B.concat(ks[p1, p1](x1, x2), ks[p1, p2](x1, x2), axis=1))
    approx(k(p2(x1), x2), B.concat(ks[p2, p1](x1, x2), ks[p2, p2](x1, x2), axis=1))
    approx(k(x1, p1(x2)), B.concat(ks[p1, p1](x1, x2), ks[p2, p1](x1, x2), axis=0))
    approx(k(x1, p2(x2)), B.concat(ks[p1, p2](x1, x2), ks[p2, p2](x1, x2), axis=0))
    with pytest.raises(ValueError):
        k.elwise(x1, p2(x3))
    with pytest.raises(ValueError):
        k.elwise(p1(x1), x3)

    # `FDD` versus `FDD`:
    approx(k(p1(x1), p1(x2)), ks[p1](x1, x2))
    approx(k(p1(x1), p2(x2)), ks[p1, p2](x1, x2))
    approx(k.elwise(p1(x1), p1(x3)), ks[p1].elwise(x1, x3))
    approx(k.elwise(p1(x1), p2(x3)), ks[p1, p2].elwise(x1, x3))

    # `MultiInput` versus input:
    approx(
        k(MultiInput(p2(x1), p1(x2)), x1),
        B.concat2d(
            [ks[p2, p1](x1, x1), ks[p2, p2](x1, x1)],
            [ks[p1, p1](x2, x1), ks[p1, p2](x2, x1)],
        ),
    )
    approx(
        k(x1, MultiInput(p2(x1), p1(x2))),
        B.concat2d(
            [ks[p1, p2](x1, x1), ks[p1, p1](x1, x2)],
            [ks[p2, p2](x1, x1), ks[p2, p1](x1, x2)],
        ),
    )
    with pytest.raises(ValueError):
        k.elwise(MultiInput(p2(x1), p1(x3)), p2(x1))
    with pytest.raises(ValueError):
        k.elwise(p2(x1), MultiInput(p2(x1), p1(x3)))

    # `MultiInput` versus `FDD`:
    approx(
        k(MultiInput(p2(x1), p1(x2)), p2(x1)),
        B.concat(ks[p2, p2](x1, x1), ks[p1, p2](x2, x1), axis=0),
    )
    approx(
        k(p2(x1), MultiInput(p2(x1), p1(x2))),
        B.concat(ks[p2, p2](x1, x1), ks[p2, p1](x1, x2), axis=1),
    )
    with pytest.raises(ValueError):
        k.elwise(MultiInput(p2(x1), p1(x3)), p2(x1))
    with pytest.raises(ValueError):
        k.elwise(p2(x1), MultiInput(p2(x1), p1(x3)))

    # `MultiInput` versus `MultiInput`:
    approx(
        k(MultiInput(p2(x1), p1(x2)), MultiInput(p2(x1))),
        B.concat(ks[p2, p2](x1, x1), ks[p1, p2](x2, x1), axis=0),
    )
    with pytest.raises(ValueError):
        k.elwise(MultiInput(p2(x1), p1(x3)), MultiInput(p2(x1)))
    approx(
        k.elwise(MultiInput(p2(x1), p1(x3)), MultiInput(p2(x1), p1(x3))),
        B.concat(ks[p2, p2].elwise(x1, x1), ks[p1, p1].elwise(x3, x3), axis=0),
    )



from setuptools import find_packages, setup

requirements = [
    "numpy>=1.16",
    "fdm",
    "algebra>=1",
    "plum-dispatch>=1",
    "backends>=1",
    "backends-matrix>=1",
]

setup(
    packages=find_packages(exclude=["docs"]),
    python_requires=">=3.6",
    install_requires=requirements,
    include_package_data=True,
)



import logging

from lab import B
from plum import Dispatcher

from . import PromisedFDD as FDD
from .input import MultiInput
from .mean import Mean

__all__ = ["MultiOutputMean"]

_dispatch = Dispatcher()

log = logging.getLogger(__name__)


class MultiOutputMean(Mean):
    """A generic multi-output mean.

    Args:
        measure (:class:`.measure.Measure`): Measure to take the means from.
        *ps (:class:`.graph.GP`): Processes that make up the multi-valued process.

    Attributes:
        measure (:class:`.measure.Measure`): Measure to take the means from.
        ps (tuple[:class:`.graph.GP`]): Processes that make up the multi-valued process.
    """

    def __init__(self, measure, *ps):
        self.measure = measure
        self.ps = ps

    @_dispatch
    def __call__(self, x: B.Numeric):
        return self(MultiInput(*(p(x) for p in self.ps)))

    @_dispatch
    def __call__(self, x: FDD):
        return self.measure.means[x.p](x.x)

    @_dispatch
    def __call__(self, x: MultiInput):
        return B.concat(*[self(xi) for xi in x.get()], axis=0)

    def render(self, formatter):
        ms = [str(self.measure.means[p]) for p in self.ps]
        return "MultiOutputMean({})".format(", ".join(ms))



from typing import Union
import logging

# noinspection PyUnresolvedReferences
import lab.tensorflow
from algebra import Element
from matrix import AbstractMatrix
from plum import Dispatcher
from tensorflow import Tensor, Variable

# noinspection PyUnresolvedReferences
from . import *  # pragma: no cover
from .random import Random

log = logging.getLogger(__name__)

_dispatch = Dispatcher()

# Save original methods.
__mul__tensor = Tensor.__mul__
__mul__variable = Variable.__mul__

__add__tensor = Tensor.__add__
__add__variable = Variable.__add__


@_dispatch
def __mul__(self: Tensor, other):
    return __mul__tensor(self, other)


@_dispatch
def __mul__(self: Variable, other):
    return __add__variable(self, other)


@_dispatch.multi(
    (Tensor, Union[Element, Random, AbstractMatrix]),
    (Variable, Union[Element, Random, AbstractMatrix]),
)
def __mul__(
    self: Union[Tensor, Variable], other: Union[Element, Random, AbstractMatrix]
):
    return other.__rmul__(self)


@_dispatch
def __add__(self: Tensor, other):
    return __add__tensor(self, other)


@_dispatch
def __add__(self: Variable, other):
    return __add__variable(self, other)


@_dispatch.multi(
    (Tensor, Union[Element, Random, AbstractMatrix]),
    (Variable, Union[Element, Random, AbstractMatrix]),
)
def __add__(
    self: Union[Tensor, Variable], other: Union[Element, Random, AbstractMatrix]
):
    return other.__radd__(self)


# Assign dispatchable methods.
Tensor.__mul__ = __mul__
Variable.__mul__ = __mul__

Tensor.__add__ = __add__
Variable.__add__ = __add__


import logging

from lab import B
from plum import Dispatcher, Union

from . import PromisedFDD as FDD
from .input import MultiInput, Input
from .kernel import Kernel

__all__ = ["MultiOutputKernel"]

_dispatch = Dispatcher()

log = logging.getLogger(__name__)


class MultiOutputKernel(Kernel):
    """A generic multi-output kernel.

    Args:
        measure (:class:`.measure.Measure`): Measure to take the kernels from.
        *ps (:class:`.measure.GP`): Processes that make up the multi-valued process.

    Attributes:
        measure (:class:`.measure.Measure`): Measure to take the kernels from.
        ps (tuple[:class:`.measure.GP`]): Processes that make up the multi-valued
            process.
    """

    def __init__(self, measure, *ps):
        self.measure = measure
        self.ps = ps

    # No `FDD` nor `MultiInput`.

    @_dispatch
    def __call__(self, x: Union[B.Numeric, Input], y: Union[B.Numeric, Input]):
        return self(
            MultiInput(*(p(x) for p in self.ps)), MultiInput(*(p(y) for p in self.ps))
        )

    # One `FDD`.

    @_dispatch
    def __call__(self, x: FDD, y: Union[B.Numeric, Input]):
        return self(MultiInput(x), MultiInput(*(p(y) for p in self.ps)))

    @_dispatch
    def __call__(self, x: Union[B.Numeric, Input], y: FDD):
        return self(MultiInput(*(p(x) for p in self.ps)), MultiInput(y))

    # Two `FDD`s.

    @_dispatch
    def __call__(self, x: FDD, y: FDD):
        return self.measure.kernels[x.p, y.p](x.x, y.x)

    # One `MultiInput`.

    @_dispatch
    def __call__(self, x: MultiInput, y: FDD):
        return self(x, MultiInput(y))

    @_dispatch
    def __call__(self, x: MultiInput, y: Union[B.Numeric, Input]):
        return self(x, MultiInput(*(p(y) for p in self.ps)))

    @_dispatch
    def __call__(self, x: FDD, y: MultiInput):
        return self(MultiInput(x), y)

    @_dispatch
    def __call__(self, x: Union[B.Numeric, Input], y: MultiInput):
        return self(MultiInput(*(p(x) for p in self.ps)), y)

    # Two `MultiInput`s.

    @_dispatch
    def __call__(self, x: MultiInput, y: MultiInput):
        return B.block(*[[self(xi, yi) for yi in y.get()] for xi in x.get()])

    # No `FDD` nor `MultiInput`.

    @_dispatch
    def elwise(self, x: Union[B.Numeric, Input], y: Union[B.Numeric, Input]):
        return self.elwise(
            MultiInput(*(p(x) for p in self.ps)), MultiInput(*(p(y) for p in self.ps))
        )

    # One `FDD`.

    @_dispatch
    def elwise(self, x: FDD, y: Union[B.Numeric, Input]):
        raise ValueError(
            "Unclear combination of arguments given to MultiOutputKernel.elwise."
        )

    @_dispatch
    def elwise(self, x: Union[B.Numeric, Input], y: FDD):
        raise ValueError(
            "Unclear combination of arguments given to " "MultiOutputKernel.elwise."
        )

    # Two `FDD`s.

    @_dispatch
    def elwise(self, x: FDD, y: FDD):
        return self.measure.kernels[x.p, y.p].elwise(x.x, y.x)

    # One `MultiInput`.

    @_dispatch(precedence=1)
    def elwise(self, x: MultiInput, y: Union[B.Numeric, Input, FDD]):
        raise ValueError(
            "Unclear combination of arguments given to MultiOutputKernel.elwise."
        )

    @_dispatch(precedence=1)
    def elwise(self, x: Union[B.Numeric, Input, FDD], y: MultiInput):
        raise ValueError(
            "Unclear combination of arguments given to MultiOutputKernel.elwise."
        )

    # Two `MultiInput`s.

    @_dispatch
    def elwise(self, x: MultiInput, y: MultiInput):
        if len(x.get()) != len(y.get()):
            raise ValueError(
                "MultiOutputKernel.elwise must be called with similarly sized "
                "MultiInputs."
            )
        return B.concat(
            *[self.elwise(xi, yi) for xi, yi in zip(x.get(), y.get())], axis=0
        )

    def render(self, formatter):
        ks = [str(self.measure.kernels[p]) for p in self.ps]
        return "MultiOutputKernel({})".format(", ".join(ks))
